@@ -23,11 +23,15 @@
 #include "list.h"
 #include "dwarves.h"
 #include "dutil.h"
-#include "strings.h"
+#include "pahole_strings.h"
 #include "hash.h"
 
 struct strings *strings;
 
+#ifndef DW_AT_alignment
+#define DW_AT_alignment 0x88
+#endif
+
 #ifndef DW_AT_GNU_vector
 #define DW_AT_GNU_vector 0x2107
 #endif
@@ -37,6 +41,15 @@ struct strings *strings;
 #define DW_TAG_GNU_call_site_parameter 0x410a
 #endif
 
+#ifndef DW_TAG_call_site
+#define DW_TAG_call_site 0x48
+#define DW_TAG_call_site_parameter 0x49
+#endif
+
+#ifndef DW_FORM_implicit_const
+#define DW_FORM_implicit_const 0x21
+#endif
+
 #define hashtags__fn(key) hash_64(key, HASHTAGS__BITS)
 
 bool no_bitfield_type_recode = true;
@@ -89,7 +102,7 @@ static void dwarf_tag__set_spec(struct dwarf_tag *dtag, dwarf_off_ref spec)
 	*(dwarf_off_ref *)(dtag + 1) = spec;
 }
 
-#define HASHTAGS__BITS 8
+#define HASHTAGS__BITS 15
 #define HASHTAGS__SIZE (1UL << HASHTAGS__BITS)
 
 #define obstack_chunk_alloc malloc
@@ -244,6 +257,7 @@ static uint64_t attr_numeric(Dwarf_Die *die, uint32_t name)
 			return addr;
 	}
 		break;
+	case DW_FORM_implicit_const:
 	case DW_FORM_data1:
 	case DW_FORM_data2:
 	case DW_FORM_data4:
@@ -286,15 +300,12 @@ static uint64_t dwarf_expr(const uint8_t *expr, uint32_t len __unused)
 	return UINT64_MAX;
 }
 
-static Dwarf_Off attr_offset(Dwarf_Die *die, const uint32_t name)
+static Dwarf_Off __attr_offset(Dwarf_Attribute *attr)
 {
-	Dwarf_Attribute attr;
 	Dwarf_Block block;
 
-	if (dwarf_attr(die, name, &attr) == NULL)
-		return 0;
-
-	switch (dwarf_whatform(&attr)) {
+	switch (dwarf_whatform(attr)) {
+	case DW_FORM_implicit_const:
 	case DW_FORM_data1:
 	case DW_FORM_data2:
 	case DW_FORM_data4:
@@ -302,18 +313,28 @@ static Dwarf_Off attr_offset(Dwarf_Die *die, const uint32_t name)
 	case DW_FORM_sdata:
 	case DW_FORM_udata: {
 		Dwarf_Word value;
-		if (dwarf_formudata(&attr, &value) == 0)
+		if (dwarf_formudata(attr, &value) == 0)
 			return value;
 		break;
 	}
 	default:
-		if (dwarf_formblock(&attr, &block) == 0)
+		if (dwarf_formblock(attr, &block) == 0)
 			return dwarf_expr(block.data, block.length);
 	}
 
 	return 0;
 }
 
+static Dwarf_Off attr_offset(Dwarf_Die *die, const uint32_t name)
+{
+	Dwarf_Attribute attr;
+
+	if (dwarf_attr(die, name, &attr) == NULL)
+		return 0;
+
+	return __attr_offset(&attr);
+}
+
 static const char *attr_string(Dwarf_Die *die, uint32_t name)
 {
 	Dwarf_Attribute attr;
@@ -472,6 +493,20 @@ static struct array_type *array_type__new(Dwarf_Die *die, struct cu *cu)
 	return at;
 }
 
+static struct string_type *string_type__new(Dwarf_Die *die, struct cu *cu)
+{
+	struct string_type *st = tag__alloc(cu, sizeof(*st));
+
+	if (st != NULL) {
+		tag__init(&st->tag, cu, die);
+		st->nr_entries = attr_numeric(die, DW_AT_byte_size);
+		if (st->nr_entries == 0)
+			st->nr_entries = 1;
+	}
+
+	return st;
+}
+
 static void namespace__init(struct namespace *namespace, Dwarf_Die *die,
 			    struct cu *cu)
 {
@@ -496,7 +531,7 @@ static struct namespace *namespace__new(Dwarf_Die *die, struct cu *cu)
 static void type__init(struct type *type, Dwarf_Die *die, struct cu *cu)
 {
 	namespace__init(&type->namespace, die, cu);
-	INIT_LIST_HEAD(&type->node);
+	__type__init(type);
 	type->size		 = attr_numeric(die, DW_AT_byte_size);
 	type->alignment		 = attr_numeric(die, DW_AT_alignment);
 	type->declaration	 = attr_numeric(die, DW_AT_declaration);
@@ -576,7 +611,15 @@ const char *variable__scope_str(const struct variable *var)
 
 static struct variable *variable__new(Dwarf_Die *die, struct cu *cu)
 {
-	struct variable *var = tag__alloc(cu, sizeof(*var));
+	struct variable *var;
+	bool has_specification;
+
+	has_specification = dwarf_hasattr(die, DW_AT_specification);
+	if (has_specification) {
+		var = tag__alloc_with_spec(cu, sizeof(*var));
+	} else {
+		var = tag__alloc(cu, sizeof(*var));
+	}
 
 	if (var != NULL) {
 		tag__init(&var->ip.tag, cu, die);
@@ -589,6 +632,10 @@ static struct variable *variable__new(Dwarf_Die *die, struct cu *cu)
 		var->ip.addr = 0;
 		if (!var->declaration && cu->has_addr_info)
 			var->scope = dwarf__location(die, &var->ip.addr, &var->location);
+		if (has_specification) {
+			dwarf_tag__set_spec(var->ip.tag.priv,
+					    attr_type(die, DW_AT_specification));
+		}
 	}
 
 	return var;
@@ -734,25 +781,41 @@ static struct class_member *class_member__new(Dwarf_Die *die, struct cu *cu,
 	if (member != NULL) {
 		tag__init(&member->tag, cu, die);
 		member->name = strings__add(strings, attr_string(die, DW_AT_name));
-		member->is_static   = !in_union && !dwarf_hasattr(die, DW_AT_data_member_location);
 		member->const_value = attr_numeric(die, DW_AT_const_value);
 		member->alignment = attr_numeric(die, DW_AT_alignment);
-		member->byte_offset = attr_offset(die, DW_AT_data_member_location);
-		/*
-		 * Bit offset calculated here is valid only for byte-aligned
-		 * fields. For bitfields on little-endian archs we need to
-		 * adjust them taking into account byte size of the field,
-		 * which might not be yet known. So we'll re-calculate bit
-		 * offset later, in class_member__cache_byte_size.
-		 */
-		member->bit_offset = member->byte_offset * 8;
+
+		Dwarf_Attribute attr;
+
+		member->has_bit_offset = dwarf_attr(die, DW_AT_data_bit_offset, &attr) != NULL;
+
+		if (member->has_bit_offset) {
+			member->bit_offset = __attr_offset(&attr);
+			// byte_offset and bitfield_offset will be recalculated later, when
+			// we discover the size of this bitfield base type.
+		} else {
+			if (dwarf_attr(die, DW_AT_data_member_location, &attr) != NULL) {
+				member->byte_offset = __attr_offset(&attr);
+			} else {
+				member->is_static = !in_union;
+			}
+
+			/*
+			 * Bit offset calculated here is valid only for byte-aligned
+			 * fields. For bitfields on little-endian archs we need to
+			 * adjust them taking into account byte size of the field,
+			 * which might not be yet known. So we'll re-calculate bit
+			 * offset later, in class_member__cache_byte_size.
+			 */
+			member->bit_offset = member->byte_offset * 8;
+			member->bitfield_offset = attr_numeric(die, DW_AT_bit_offset);
+		}
+
 		/*
 		 * If DW_AT_byte_size is not present, byte size will be
 		 * determined later in class_member__cache_byte_size using
 		 * base integer/enum type
 		 */
 		member->byte_size = attr_numeric(die, DW_AT_byte_size);
-		member->bitfield_offset = attr_numeric(die, DW_AT_bit_offset);
 		member->bitfield_size = attr_numeric(die, DW_AT_bit_size);
 		member->bit_hole = 0;
 		member->bitfield_end = 0;
@@ -1127,6 +1190,16 @@ out_free:
 	return NULL;
 }
 
+static struct tag *die__create_new_string_type(Dwarf_Die *die, struct cu *cu)
+{
+	struct string_type *string = string_type__new(die, cu);
+
+	if (string == NULL)
+		return NULL;
+
+	return &string->tag;
+}
+
 static struct tag *die__create_new_parameter(Dwarf_Die *die,
 					     struct ftype *ftype,
 					     struct lexblock *lexblock,
@@ -1193,6 +1266,9 @@ static struct tag *die__create_new_subroutine_type(Dwarf_Die *die,
 		uint32_t id;
 
 		switch (dwarf_tag(die)) {
+		case DW_TAG_subrange_type: // ADA stuff
+			tag__print_not_supported(dwarf_tag(die));
+			continue;
 		case DW_TAG_formal_parameter:
 			tag = die__create_new_parameter(die, ftype, NULL, cu);
 			break;
@@ -1204,6 +1280,11 @@ static struct tag *die__create_new_subroutine_type(Dwarf_Die *die,
 			if (tag == NULL)
 				goto out_delete;
 
+			if (tag == &unsupported_tag) {
+				tag__print_not_supported(dwarf_tag(die));
+				continue;
+			}
+
 			if (cu__add_tag(cu, tag, &id) < 0)
 				goto out_delete_tag;
 
@@ -1276,6 +1357,8 @@ static int die__process_class(Dwarf_Die *die, struct type *class,
 
 	do {
 		switch (dwarf_tag(die)) {
+		case DW_TAG_subrange_type: // XXX: ADA stuff, its a type tho, will have other entries referencing it...
+		case DW_TAG_variant_part: // XXX: Rust stuff
 #ifdef STB_GNU_UNIQUE
 		case DW_TAG_GNU_formal_parameter_pack:
 		case DW_TAG_GNU_template_parameter_pack:
@@ -1321,6 +1404,11 @@ static int die__process_class(Dwarf_Die *die, struct type *class,
 			if (tag == NULL)
 				return -ENOMEM;
 
+			if (tag == &unsupported_tag) {
+				tag__print_not_supported(dwarf_tag(die));
+				continue;
+			}
+
 			uint32_t id;
 
 			if (cu__table_add_tag(cu, tag, &id) < 0) {
@@ -1356,6 +1444,11 @@ static int die__process_namespace(Dwarf_Die *die, struct namespace *namespace,
 		if (tag == NULL)
 			goto out_enomem;
 
+		if (tag == &unsupported_tag) {
+			tag__print_not_supported(dwarf_tag(die));
+			continue;
+		}
+
 		uint32_t id;
 		if (cu__table_add_tag(cu, tag, &id) < 0)
 			goto out_delete_tag;
@@ -1411,6 +1504,8 @@ static int die__process_inline_expansion(Dwarf_Die *die, struct lexblock *lexblo
 		uint32_t id;
 
 		switch (dwarf_tag(die)) {
+		case DW_TAG_call_site:
+		case DW_TAG_call_site_parameter:
 		case DW_TAG_GNU_call_site:
 		case DW_TAG_GNU_call_site_parameter:
 			/*
@@ -1451,8 +1546,10 @@ static int die__process_inline_expansion(Dwarf_Die *die, struct lexblock *lexblo
 			if (tag == NULL)
 				goto out_enomem;
 
-			if (tag == &unsupported_tag)
+			if (tag == &unsupported_tag) {
+				tag__print_not_supported(dwarf_tag(die));
 				continue;
+			}
 
 			if (cu__add_tag(cu, tag, &id) < 0)
 				goto out_delete_tag;
@@ -1510,6 +1607,8 @@ static int die__process_function(Dwarf_Die *die, struct ftype *ftype,
 		uint32_t id;
 
 		switch (dwarf_tag(die)) {
+		case DW_TAG_call_site:
+		case DW_TAG_call_site_parameter:
 		case DW_TAG_GNU_call_site:
 		case DW_TAG_GNU_call_site_parameter:
 			/*
@@ -1568,8 +1667,10 @@ static int die__process_function(Dwarf_Die *die, struct ftype *ftype,
 			if (tag == NULL)
 				goto out_enomem;
 
-			if (tag == &unsupported_tag)
+			if (tag == &unsupported_tag) {
+				tag__print_not_supported(dwarf_tag(die));
 				continue;
+			}
 
 			if (cu__add_tag(cu, tag, &id) < 0)
 				goto out_delete_tag;
@@ -1615,8 +1716,12 @@ static struct tag *__die__process_tag(Dwarf_Die *die, struct cu *cu,
 	struct tag *tag;
 
 	switch (dwarf_tag(die)) {
+	case DW_TAG_imported_unit:
+		return NULL; // We don't support imported units yet, so to avoid segfaults
 	case DW_TAG_array_type:
 		tag = die__create_new_array(die, cu);		break;
+	case DW_TAG_string_type: // FORTRAN stuff, looks like an array
+		tag = die__create_new_string_type(die, cu);	break;
 	case DW_TAG_base_type:
 		tag = die__create_new_base_type(die, cu);	break;
 	case DW_TAG_const_type:
@@ -1673,8 +1778,13 @@ static int die__process_unit(Dwarf_Die *die, struct cu *cu)
 		if (tag == NULL)
 			return -ENOMEM;
 
-		if (tag == &unsupported_tag)
+		if (tag == &unsupported_tag) {
+			// XXX special case DW_TAG_dwarf_procedure, appears when looking at a recent ~/bin/perf
+			// Investigate later how to properly support this...
+			if (dwarf_tag(die) != DW_TAG_dwarf_procedure)
+				tag__print_not_supported(dwarf_tag(die));
 			continue;
+		}
 
 		uint32_t id;
 		cu__add_tag(cu, tag, &id);
@@ -2008,6 +2118,17 @@ static int tag__recode_dwarf_type(struct tag *tag, struct cu *cu)
 		if (dtype != NULL)
 			goto out;
 		goto find_type;
+	case DW_TAG_variable: {
+		struct variable *var = tag__variable(tag);
+		dwarf_off_ref specification = dwarf_tag__spec(dtag);
+
+		if (specification.off) {
+			dtype = dwarf_cu__find_tag_by_ref(cu->priv, &specification);
+			if (dtype)
+				var->spec = tag__variable(dtype->tag);
+		}
+	}
+
 	}
 
 	if (dtag->type.off == 0) {
@@ -2075,7 +2196,7 @@ static unsigned long long dwarf_tag__orig_id(const struct tag *tag,
 static const char *dwarf__strings_ptr(const struct cu *cu __unused,
 				      strings_t s)
 {
-	return strings__ptr(strings, s);
+	return s ? strings__ptr(strings, s) : NULL;
 }
 
 struct debug_fmt_ops dwarf__ops;
@@ -2085,7 +2206,19 @@ static int die__process(Dwarf_Die *die, struct cu *cu)
 	Dwarf_Die child;
 	const uint16_t tag = dwarf_tag(die);
 
-	if (tag != DW_TAG_compile_unit && tag != DW_TAG_type_unit && tag != DW_TAG_partial_unit) {
+	if (tag == DW_TAG_partial_unit) {
+		static bool warned;
+
+		if (!warned) {
+			fprintf(stderr, "WARNING: DW_TAG_partial_unit used, some types will not be considered!\n"
+					"         Probably this was optimized using a tool like 'dwz'\n"
+					"         A future version of pahole will support this.\n");
+			warned = true;
+		}
+		return 0; // so that other units can be processed
+	}
+
+	if (tag != DW_TAG_compile_unit && tag != DW_TAG_type_unit) {
 		fprintf(stderr, "%s: DW_TAG_compile_unit, DW_TAG_type_unit or DW_TAG_partial_unit expected got %s!\n",
 			__FUNCTION__, dwarf_tag_name(tag));
 		return -EINVAL;
@@ -2154,7 +2287,7 @@ static int class_member__cache_byte_size(struct tag *tag, struct cu *cu,
 	member->bit_size = member->byte_size * 8;
 
 	/*
-	 * XXX: after all the attemps to determine byte size, we might still
+	 * XXX: after all the attempts to determine byte size, we might still
 	 * be unsuccessful, because base_type__name_to_size doesn't know about
 	 * the base_type name, so one has to add there when such base_type
 	 * isn't found. pahole will put zero on the struct output so it should
@@ -2165,24 +2298,31 @@ static int class_member__cache_byte_size(struct tag *tag, struct cu *cu,
 		return 0;
 	}
 
-	/*
-	 * For little-endian architectures, DWARF data emitted by gcc/clang
-	 * specifies bitfield offset as an offset from the highest-order bit
-	 * of an underlying integral type (e.g., int) to a highest-order bit
-	 * of a bitfield. E.g., for bitfield taking first 5 bits of int-backed
-	 * bitfield, bit offset will be 27 (sizeof(int) - 0 offset - 5 bit
-	 * size), which is very counter-intuitive and isn't a natural
-	 * extension of byte offset, which on little-endian points to
-	 * lowest-order byte. So here we re-adjust bitfield offset to be an
-	 * offset from lowest-order bit of underlying integral type to
-	 * a lowest-order bit of a bitfield. This makes bitfield offset
-	 * a natural extension of byte offset for bitfields and is uniform
-	 * with how big-endian bit offsets work.
-	 */
-	if (cu->little_endian) {
-		member->bitfield_offset = member->bit_size - member->bitfield_offset - member->bitfield_size;
+	if (!member->has_bit_offset) {
+		/*
+		 * For little-endian architectures, DWARF data emitted by gcc/clang
+		 * specifies bitfield offset as an offset from the highest-order bit
+		 * of an underlying integral type (e.g., int) to a highest-order bit
+		 * of a bitfield. E.g., for bitfield taking first 5 bits of int-backed
+		 * bitfield, bit offset will be 27 (sizeof(int) - 0 offset - 5 bit
+		 * size), which is very counter-intuitive and isn't a natural
+		 * extension of byte offset, which on little-endian points to
+		 * lowest-order byte. So here we re-adjust bitfield offset to be an
+		 * offset from lowest-order bit of underlying integral type to
+		 * a lowest-order bit of a bitfield. This makes bitfield offset
+		 * a natural extension of byte offset for bitfields and is uniform
+		 * with how big-endian bit offsets work.
+		 */
+		if (cu->little_endian)
+			member->bitfield_offset = member->bit_size - member->bitfield_offset - member->bitfield_size;
+
+		member->bit_offset = member->byte_offset * 8 + member->bitfield_offset;
+	} else {
+		// DWARF5 has DW_AT_data_bit_offset, offset in bits from the
+		// start of the container type (struct, class, etc).
+		member->byte_offset = member->bit_offset / 8;
+		member->bitfield_offset = member->bit_offset - member->byte_offset * 8;
 	}
-	member->bit_offset = member->byte_offset * 8 + member->bitfield_offset;
 
 	/* make sure bitfield offset is non-negative */
 	if (member->bitfield_offset < 0) {
@@ -2309,11 +2449,11 @@ static int cus__load_module(struct cus *cus, struct conf_load *conf,
 {
 	Dwarf_Off off = 0, noff;
 	size_t cuhl;
-	GElf_Addr vaddr;
 	const unsigned char *build_id = NULL;
 	uint8_t pointer_size, offset_size;
 
 #ifdef HAVE_DWFL_MODULE_BUILD_ID
+	GElf_Addr vaddr;
 	int build_id_len = dwfl_module_build_id(mod, &build_id, &vaddr);
 #else
 	int build_id_len = 0;
@@ -2342,6 +2482,9 @@ static int cus__load_module(struct cus *cus, struct conf_load *conf,
 		Dwarf_Die die_mem;
 		Dwarf_Die *cu_die = dwarf_offdie(dw, off + cuhl, &die_mem);
 
+		if (cu_die == NULL)
+			break;
+
 		/*
 		 * DW_AT_name in DW_TAG_compile_unit can be NULL, first
 		 * seen in:
