@@ -33,6 +33,8 @@ static bool expand_types;
 static bool compilable_output;
 static struct type_emissions emissions;
 static uint64_t addr;
+static char *class_name;
+static char *function_name;
 
 static struct conf_fprintf conf;
 
@@ -494,6 +496,23 @@ int elf_symtabs__show(char *filenames[])
 	return EXIT_SUCCESS;
 }
 
+static enum load_steal_kind pfunct_stealer(struct cu *cu, struct conf_load *conf_load __unused)
+{
+
+	if (function_name) {
+		struct tag *tag = cu__find_function_by_name(cu, function_name);
+
+		if (tag) {
+			function__show(tag__function(tag), cu);
+			return LSK__STOP_LOADING;
+		}
+	} else if (class_name) {
+		cu_class_iterator(cu, class_name);
+	}
+
+	return LSK__DELETE;
+}
+
 /* Name and version of program.  */
 ARGP_PROGRAM_VERSION_HOOK_DEF = dwarves_print_version;
 
@@ -631,8 +650,6 @@ static const struct argp_option pfunct__options[] = {
 };
 
 static void (*formatter)(const struct fn_stats *f) = fn_stats_fmtr;
-static char *class_name;
-static char *function_name;
 static int show_total_inline_expansion_stats;
 
 static error_t pfunct__options_parser(int key, char *arg,
@@ -722,12 +739,28 @@ int main(int argc, char *argv[])
 		goto out_dwarves_exit;
 	}
 
+	if (function_name || class_name)
+		conf_load.steal = pfunct_stealer;
+
+try_sole_arg_as_function_name:
 	err = cus__load_files(cus, &conf_load, argv + remaining);
 	if (err != 0) {
+		if (function_name == NULL) {
+                        function_name = argv[remaining];
+                        if (access(function_name, R_OK) == 0) {
+                                fprintf(stderr, "pfunct: file '%s' has no %s type information.\n",
+                                                function_name, conf_load.format_path ?: "supported");
+                                goto out_dwarves_exit;
+                        }
+			conf_load.steal = pfunct_stealer;
+                        remaining = argc;
+			goto try_sole_arg_as_function_name;
+		}
 		cus__fprintf_load_files_err(cus, "pfunct", argv + remaining, err, stderr);
 		goto out_cus_delete;
 	}
 
+
 	cus__for_each_cu(cus, cu_unique_iterator, NULL, NULL);
 
 	if (addr) {
@@ -742,8 +775,6 @@ int main(int argc, char *argv[])
 		function__show(f, cu);
 	} else if (show_total_inline_expansion_stats)
 		print_total_inline_stats();
-	else if (class_name != NULL)
-		cus__for_each_cu(cus, cu_class_iterator, class_name, NULL);
 	else if (function_name != NULL || expand_types)
 		cus__for_each_cu(cus, cu_function_iterator,
 				 function_name, NULL);
